import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:url_router/url_router.dart';

class UrlRouter extends RouterDelegate<String> with ChangeNotifier, PopNavigatorRouterDelegateMixin {
  UrlRouter({String url = '/', required this.onGeneratePages, this.onPopPage, this.onChanging, this.scaffoldBuilder}) {
    _initialUrl = url;
  }

  static UrlRouteParser parser = UrlRouteParser();

  /// Enable UrlRouter.of(context) lookup
  static UrlRouter of(BuildContext context) =>
      (context.dependOnInheritedWidgetOfExactType<_InheritedRouterController>() as _InheritedRouterController).router;

  /// Should build a stack of pages, based on the current location.
  final List<Page<dynamic>> Function(UrlRouter router) onGeneratePages;

  /// Optionally provide a way for the parent to implement custom `onPopPage` logic.
  final PopPageCallback? onPopPage;

  /// Optionally invoked just prior to the location being changed.
  /// Allows a parent class to protect or redirect certain routes. The callback can return the original url to allow the location change,
  /// or return a new url to redirect. If null is returned the location change will be ignored / blocked.
  final String? Function(UrlRouter router, String newLocation)? onChanging;

  /// Wrap widgets around the Navigator object, providing scaffolding like a `SideBar`, `TitleBar` etc.
  final Widget Function(UrlRouter router, Widget navigator)? scaffoldBuilder;

  /// Set from inside the build method, allows us to avoid passing context into delegates
  late BuildContext context;

  Map<String, String> get queryParams => _getUri()?.queryParameters ?? {};

  @override
  @protected
  String? get currentConfiguration => _url;

  @override
  GlobalKey<NavigatorState>? get navigatorKey => _navKey;
  final GlobalKey<NavigatorState> _navKey = GlobalKey();

  Uri? _getUri() => Uri.tryParse(_url);

  late final String _initialUrl;
  String _url = '';
  String get url => _url;
  set url(String value) {
    if (value != _url) {
      // Allow onChanging to override the target path
      String? newUrl = onChanging?.call(this, value);
      if (newUrl != null) {
        _url = newUrl;
        notifyListeners();
      }
    }
  }

  void push(String path) {
    final segments = _getUri()?.pathSegments.toList() ?? [];
    segments.add(path);
    url = '/' + segments.join('/');
  }

  void pop() {
    final segments = _getUri()?.pathSegments.toList() ?? [];
    if (segments.length <= 1) return;
    segments.removeAt(segments.length - 1);
    url = '/' + segments.join('/');
  }

  @override
  Future<void> setInitialRoutePath(String configuration) {
    if (configuration == '/') {
      configuration = _initialUrl;
    }
    url = configuration;
    super.setInitialRoutePath(url);
    return SynchronousFuture(null);
  }

  @override
  Widget build(BuildContext context) {
    this.context = context;
    Widget content = Navigator(
      key: _navKey,
      //TODO: Figure out how to test this properly...what's it really supposed to do?
      // A page in the stack might want to handle pop itself, does this mean that it uses WillPopScope??
      onPopPage: onPopPage ??
          (route, settings) {
            if (route.didPop(settings)) {
              return true;
            }
            return false;
          },
      pages: onGeneratePages.call(this),
    );
    if (scaffoldBuilder != null) {
      content = scaffoldBuilder!.call(this, content);
    }
    return _InheritedRouterController(router: this, child: content);
  }

  @override
  SynchronousFuture<void> setNewRoutePath(configuration) {
    url = configuration;
    return SynchronousFuture(null);
  }
}

class _InheritedRouterController extends InheritedWidget {
  const _InheritedRouterController({Key? key, required Widget child, required this.router})
      : super(key: key, child: child);
  final UrlRouter router;
  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}
